import json
import uuid
from unittest.mock import Mock, patch

import pytest
from polaris.tests.helpers import mock_check_auth_success
from polaris.models import Transaction, Asset


endpoint = "/sep6/transactions/"


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep6.transaction.rdi.patch_transaction")
def test_success_update(mock_patch_transaction, client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep6,
        kind=Transaction.KIND.deposit,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    params = {"transaction": {"bank_account": "foo"}}
    response = client.patch(
        endpoint + str(transaction.id), params, content_type="application/json"
    )
    transaction.refresh_from_db()
    assert response.status_code == 200
    assert transaction.required_info_updates is None
    assert transaction.required_info_message is None
    assert transaction.status == Transaction.STATUS.pending_anchor
    mock_patch_transaction.assert_called_once()
    assert mock_patch_transaction.call_args_list[0][1].get("params") == params
    assert mock_patch_transaction.call_args_list[0][1].get("transaction") == transaction


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_not_implemented(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep6,
        kind=Transaction.KIND.deposit,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    params = {"transaction": {"bank_account": "foo"}}
    response = client.patch(
        endpoint + str(transaction.id), params, content_type="application/json"
    )
    transaction.refresh_from_db()
    assert response.status_code == 501
    assert response.json() == {"error": "not implemented"}
    assert transaction.required_info_message
    assert transaction.required_info_updates


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_bad_id(client, acc1_usd_deposit_transaction_factory):
    response = client.patch(
        endpoint + "not an id",
        {"transaction": {"bank_account": "foo"}},
        content_type="application/json",
    )
    assert response.status_code == 404


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_missing_id(client):
    response = client.patch(
        endpoint,
        {"transaction": {"bank_account": "foo"}},
        content_type="application/json",
    )
    assert response.status_code == 404


def test_no_auth(client):
    response = client.patch(endpoint + "test")
    assert response.status_code == 403


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_not_found(client):
    response = client.patch(
        endpoint + str(uuid.uuid4()),
        {"transaction": {"bank_account": "foo"}},
        content_type="application/json",
    )
    assert response.status_code == 404


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_update_not_required(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep6,
        kind=Transaction.KIND.deposit,
        status=Transaction.STATUS.pending_user_transfer_start,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"transaction": {"bank_account": "foo"}},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert "update not required" in response.json()["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_bad_info_update_column(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep6,
        kind=Transaction.KIND.deposit,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates="test",
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"transaction": {"bank_account": "foo"}},
        content_type="application/json",
    )
    assert response.status_code == 500


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_bad_update_body(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep6,
        kind=Transaction.KIND.deposit,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"transaction": {"not a listed field": True}},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert "missing bank_account" in response.json()["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_missing_update_category(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep6,
        kind=Transaction.KIND.deposit,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"first_name": "test"},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert "missing transaction" in response.json()["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_bad_category_value_type(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep6,
        kind=Transaction.KIND.deposit,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"transaction": "not a dict"},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert "invalid type" in response.json()["error"]


raise_error_integration = Mock(patch_transaction=Mock(side_effect=ValueError("test")))


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep6.transaction.rdi", raise_error_integration,
)
def test_user_defined_exception(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep6,
        kind=Transaction.KIND.deposit,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"transaction": {"bank_account": "foo"}},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert response.json()["error"] == "test"



import sys
import signal
import time
import datetime
import asyncio
from decimal import Decimal
from typing import Tuple, List, Optional, Dict
from collections import defaultdict

import django.db.transaction
from django.core.management import BaseCommand
from stellar_sdk import (
    Keypair,
    TransactionEnvelope,
    Asset,
    Claimant,
    Server,
    MuxedAccount,
)
from stellar_sdk.client.aiohttp_client import AiohttpClient
from stellar_sdk.account import Account
from stellar_sdk.exceptions import (
    BaseHorizonError,
    ConnectionError,
    NotFoundError,
    BaseRequestError,
)
from stellar_sdk.transaction_builder import TransactionBuilder
from stellar_sdk.xdr import TransactionResult, OperationType
from asgiref.sync import sync_to_async

from polaris import settings
from polaris.utils import (
    is_pending_trust,
    maybe_make_callback,
    maybe_make_callback_async,
    get_account_obj_async,
)
from polaris.integrations import (
    registered_deposit_integration as rdi,
    registered_rails_integration as rri,
    registered_fee_func,
    calculate_fee,
)
from polaris.models import Transaction
from polaris.utils import getLogger, make_memo

logger = getLogger(__name__)

TERMINATE = False
"""
SIGINT and SIGTERM signals to this process set TERMINATE to True,
and once all pending tasks complete, the process exits.
Only relevant if the --loop option is specified.
"""

DEFAULT_INTERVAL = 10
"""
The default amount of time to sleep before querying for transactions again
Only used if the --loop option is specified.
"""


class Command(BaseCommand):
    """
    The process_pending_deposits command handler.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    @staticmethod
    def exit_gracefully(*_):  # pragma: no cover
        logger.info("Exiting process_pending_deposits...")
        module = sys.modules[__name__]
        module.TERMINATE = True

    @staticmethod
    def sleep(seconds):  # pragma: no cover
        for _ in range(seconds):
            if TERMINATE:
                break
            time.sleep(1)

    def add_arguments(self, parser):  # pragma: no cover
        parser.add_argument(
            "--loop",
            action="store_true",
            help="Continually restart command after a specified number of seconds.",
        )
        parser.add_argument(
            "--interval",
            "-i",
            type=int,
            help="The number of seconds to wait before restarting command. "
            "Defaults to {}.".format(DEFAULT_INTERVAL),
        )

    def handle(self, *_args, **options):  # pragma: no cover
        """
        The entrypoint for the functionality implemented in this file.

        Calls process_deposits(), and if the --loop option is used, does so
        periodically after sleeping for the number of seconds specified by
        --interval.
        """
        if options.get("loop"):
            while True:
                if TERMINATE:
                    break
                asyncio.run(self.process_deposits())
                self.sleep(options.get("interval") or DEFAULT_INTERVAL)
        else:
            asyncio.run(self.process_deposits())

    @classmethod
    async def process_deposits(cls):
        """
        The entry-point for the command's functionality. Queries for transactions in three states:
        - Transactions that have become available in the anchor off-chain account
        - Multi-sig transactions that have all signatures collected
        - Transactions waiting for the client application to establish a trustline for the request asset

        All transactions retreived have ``Transaction.pending_execution_attempt`` set to True to ensure
        they are not retreived by other invocation of this command.

        Tasks are queued for execution on the event loop for every transaction retreived. All tasks
        ensure that ``Transaction.pending_execution_attempt`` is set to False by the end of execution.
        """
        ready_transactions = await sync_to_async(PendingDeposits.get_ready_deposits)()
        ready_multisig_transactions = await sync_to_async(
            PendingDeposits.get_ready_multisig_deposits
        )()
        pending_trust_transactions = await sync_to_async(
            PendingDeposits.get_pending_trust_transactions
        )()
        locks = {
            "source_accounts": defaultdict(asyncio.Lock),
            "destination_accounts": defaultdict(asyncio.Lock),
        }
        async with Server(settings.HORIZON_URI, client=AiohttpClient()) as server:
            results = await asyncio.gather(
                *[
                    PendingDeposits.process_deposit(t, server, locks)
                    for t in ready_transactions
                ],
                *[
                    PendingDeposits.handle_submit(t, server, locks)
                    for t in ready_multisig_transactions
                ],
                *[
                    PendingDeposits.check_trustline(t, server, locks)
                    for t in pending_trust_transactions
                ],
                return_exceptions=True,
            )
        await cls.handle_unexpected_exceptions(
            results,
            ready_transactions
            + ready_multisig_transactions
            + pending_trust_transactions,
        )

    @classmethod
    async def handle_unexpected_exceptions(cls, results, processed_transactions):
        for i, transaction in enumerate(processed_transactions):
            if results[i] is None:  # no exeption raised
                continue
            try:
                raise results[i]
            except type(results[i]):
                logger.exception(
                    f"An unexpected exception occured while processing transaction {transaction.id}"
                )
                transaction.status_message = str(results[i])
                transaction.status = Transaction.STATUS.error
                transaction.pending_execution_attempt = False
                await sync_to_async(transaction.save)()
                await maybe_make_callback_async(transaction)


class PendingDeposits:
    @classmethod
    async def process_deposit(cls, transaction, server, locks):
        """
        Evaluate `transaction` and determine if it is ready for submission to the
        Stellar Network. The transaction could be in one of the following states:

        - The destination account does not exist or does not have a trustline to
        the asset and the initiating client application does not support claimable
        balances.

            If the account exists, or after it has been created, the transaction is
            placed in the `pending_trust` status. If the account doesn't exist and
            the distribution account requires multiple signatures, Polaris requests
            a channel account from the Anchor and submits a create account operation
            using the channel account as the source instead of the distribution
            account.

        - The distribution account that will submit the transaction requires
        multiple signatures that have not been collected.

            In this case Transaction.pending_signatures is set to True, and the
            the anchor is expected collect signatures and set the column back to
            False.

        - The transaction is ready to be submitted

            In this case, the transaction is submitted to the Stellar Network and
            updated as complete after success.
        """
        logger.info(f"processing transaction {transaction.id}")
        if await cls.requires_trustline(transaction, server, locks):
            return

        logger.info(f"transaction {transaction.id} has the appropriate trustline")
        try:
            requires_multisig = await cls.requires_multisig(transaction)
        except NotFoundError:
            await sync_to_async(cls.handle_error)(
                transaction,
                f"{transaction.asset.code} distribution account "
                f"{transaction.asset.distribution_account} does not exist",
            )
            await maybe_make_callback_async(transaction)
            return
        except ConnectionError:
            await sync_to_async(cls.handle_error)(
                transaction,
                f"Unable to connect to horizon to fetch {transaction.asset.code} "
                "distribution account signers",
            )
            await maybe_make_callback_async(transaction)
            return
        if requires_multisig:
            logger.info(f"transaction {transaction.id} requires multiple signatures")
            await cls.save_as_pending_signatures(transaction, server)
            return
        logger.debug(f"transaction {transaction.id} does not require multisig")

        await cls.handle_submit(transaction, server, locks)

    @classmethod
    def get_ready_deposits(cls) -> List[Transaction]:
        """
        Polaris' API server processes deposit request and places the associated Transaction
        object in the `pending_user_transfer_start` status when all information necessary to
        submit the payment operation on Stellar has been collected.

        This function queries for these transaction, in addition to the Transaction objects
        that have been identified as pending external rails, and passes them to the
        DepositIntegration.poll_pending_deposits() integration function. Anchors return the
        transactions that are now available in their off-chain account and therefore ready
        for submission to the Stellar Network. Finally, this function performs various
        validations to ensure the transaction is truly ready and returns them.
        """
        pending_deposits = (
            Transaction.objects.filter(
                status__in=[
                    Transaction.STATUS.pending_user_transfer_start,
                    Transaction.STATUS.pending_external,
                ],
                kind__in=[
                    Transaction.KIND.deposit,
                    getattr(Transaction.KIND, "deposit-exchange"),
                ],
                pending_execution_attempt=False,
            )
            .select_related("asset")
            .select_for_update()
        )
        with django.db.transaction.atomic():
            ready_transactions = rri.poll_pending_deposits(pending_deposits)
            Transaction.objects.filter(
                id__in=[t.id for t in ready_transactions]
            ).update(pending_execution_attempt=True)
        verified_ready_transactions = []
        for transaction in ready_transactions:
            if not transaction.amount_fee or not transaction.amount_out:
                if transaction.quote:
                    logger.error(
                        f"transaction {transaction.id} uses a quote but was returned "
                        "from poll_pending_deposits() without amount_fee or amount_out "
                        "assigned, skipping"
                    )
                    continue
                logger.warning(
                    f"transaction {transaction.id} was returned from "
                    f"poll_pending_deposits() without Transaction.amount_fee or "
                    f"Transaction.amount_out assigned. Future Polaris "
                    "releases will not calculate fees and delivered amounts"
                )
            # refresh from DB to pull pending_execution_attempt value and to ensure invalid
            # values were not assigned to the transaction in rri.poll_pending_deposits()
            asset = transaction.asset
            quote = transaction.quote
            transaction.refresh_from_db()
            transaction.asset = asset
            transaction.quote = quote
            if transaction.kind not in [
                transaction.KIND.deposit,
                getattr(transaction.KIND, "deposit-exchange"),
            ]:
                cls.handle_error(
                    transaction,
                    "poll_pending_deposits() returned a non-deposit transaction",
                )
                maybe_make_callback(transaction)
                continue
            if transaction.amount_in is None:
                cls.handle_error(
                    transaction,
                    "poll_pending_deposits() did not assign a value to the "
                    "amount_in field of a Transaction object returned",
                )
                maybe_make_callback(transaction)
                continue
            elif transaction.amount_fee is None:
                if registered_fee_func is calculate_fee:
                    try:
                        transaction.amount_fee = calculate_fee(
                            fee_params={
                                "amount": transaction.amount_in,
                                "operation": settings.OPERATION_DEPOSIT,
                                "asset_code": transaction.asset.code,
                            }
                        )
                    except ValueError:
                        transaction.amount_fee = Decimal(0)
                else:
                    transaction.amount_fee = Decimal(0)
                transaction.save()
            verified_ready_transactions.append(transaction)
        return verified_ready_transactions

    @staticmethod
    def get_pending_trust_transactions():
        """
        If the destination account does not have a trustline to the requested
        asset and the client application that initiated the request does not
        support claimable balances, Polaris places the transaction in the
        `pending_trust` status.

        This function retrieves those transactions and ensure other processes
        do not retreive the same transactions for processing by updating
        Transaction.pending_execution_attempt to True.

        The returned transactions will be submitted if their destination
        accounts now have a trustline to the asset.
        """
        with django.db.transaction.atomic():
            transactions = list(
                Transaction.objects.filter(
                    kind=Transaction.KIND.deposit,
                    status=Transaction.STATUS.pending_trust,
                    pending_execution_attempt=False,
                )
                .select_related("asset")
                .select_for_update()
            )
            Transaction.objects.filter(id__in=[t.id for t in transactions]).update(
                pending_execution_attempt=True
            )
            return transactions

    @staticmethod
    def get_ready_multisig_deposits():
        """
        If the anchor's distribution account requires multiple signatures before
        submitting to Stellar, Polaris generates the envelope and updates
        Transaction.pending_signatures to True.

        Polaris then expects the anchor to collect the necessary signatures and
        set Transaction.pending_signatures back to False. This function checks if
        any transaction is in this state and returns it for submission to the
        Stellar Network.

        Multisig transactions are therefore identified by a non-null envelope_xdr
        column and a 'pending_anchor' status. The status check is important
        because all successfully submitted transactions have their envelope_xdr
        column set after submission and status set to 'completed'.
        """
        with django.db.transaction.atomic():
            multisig_transactions = list(
                Transaction.objects.filter(
                    kind=Transaction.KIND.deposit,
                    status=Transaction.STATUS.pending_anchor,
                    pending_signatures=False,
                    envelope_xdr__isnull=False,
                    pending_execution_attempt=False,
                )
                .select_related("asset")
                .select_for_update()
            )
            Transaction.objects.filter(
                id__in=[t.id for t in multisig_transactions]
            ).update(pending_execution_attempt=True)
            for t in multisig_transactions:
                logger.debug(
                    f"Detected multisig transaction {t.id} is ready to be submitted"
                )
            return multisig_transactions

    @classmethod
    async def get_or_create_destination_account(
        cls, transaction: Transaction, server: Server, locks: Dict
    ) -> Tuple[Account, bool]:
        """
        Returns:
            Account: The account found or created for the Transaction
            bool: True if trustline doesn't exist, False otherwise.

        If the account doesn't exist, Polaris must create the account using an account provided by the
        anchor. Polaris can use the distribution account of the anchored asset or a channel account if
        the asset's distribution account requires non-master signatures.

        If the transacted asset's distribution account does require non-master signatures,
        DepositIntegration.create_channel_account() will be called. See the function docstring for more
        info.

        On failure to create the destination account, a RuntimeError exception is raised.
        """
        logger.debug(
            f"requesting lock to get or create destination account for transaction {transaction.id}"
        )
        async with locks["destination_accounts"][transaction.to_address]:
            logger.debug(
                f"got lock to get or create destination account for transaction {transaction.id}"
            )
            if transaction.to_address.startswith("M"):
                destination_account = MuxedAccount.from_account(
                    transaction.to_address
                ).account_id
            else:
                destination_account = transaction.to_address
            try:
                account, json_resp = await get_account_obj_async(
                    Keypair.from_public_key(destination_account), server
                )
                logger.debug(f"account for transaction {transaction.id} exists")
                return account, is_pending_trust(transaction, json_resp)
            except RuntimeError:  # account does not exist
                logger.debug(f"account for transaction {transaction.id} does not exist")
                try:
                    requires_multisig = await PendingDeposits.requires_multisig(
                        transaction
                    )
                except NotFoundError:
                    logger.error(
                        f"{transaction.asset.code} distribution account "
                        f"{transaction.asset.distribution_account} does not exist"
                    )
                    raise RuntimeError("the distribution account does not exist")
                if requires_multisig:
                    source_account_kp = await sync_to_async(cls.get_channel_keypair)(
                        transaction
                    )
                else:
                    source_account_kp = Keypair.from_secret(
                        transaction.asset.distribution_seed
                    )

                logger.debug(
                    f"requesting lock to create account for transaction {transaction.id}"
                )
                async with locks["source_accounts"][
                    transaction.asset.distribution_account
                ]:
                    logger.debug(
                        f"locked for transaction {transaction.id} to create destination account"
                    )
                    source_account, _ = await get_account_obj_async(
                        source_account_kp, server
                    )
                    builder = TransactionBuilder(
                        source_account=source_account,
                        network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
                        # this transaction contains one operation so base_fee will be multiplied by 1
                        base_fee=settings.MAX_TRANSACTION_FEE_STROOPS
                        or await server.fetch_base_fee(),
                    )
                    transaction_envelope = builder.append_create_account_op(
                        destination=transaction.to_address,
                        starting_balance=settings.ACCOUNT_STARTING_BALANCE,
                    ).build()
                    transaction_envelope.sign(source_account_kp)

                    try:
                        await server.submit_transaction(transaction_envelope)
                    except BaseHorizonError as e:  # pragma: no cover
                        raise RuntimeError(
                            "Horizon error when submitting create account "
                            f"to horizon: {e.message}"
                        )
                    logger.debug(
                        f"unlocking after creating destination account for transaction {transaction.id}"
                    )

                account, _ = await get_account_obj_async(
                    Keypair.from_public_key(transaction.to_address), server
                )
                return account, True
            except BaseHorizonError as e:
                raise RuntimeError(
                    f"Horizon error when loading stellar account: {e.message}"
                )
            except ConnectionError:
                raise RuntimeError("Failed to connect to Horizon")

    @classmethod
    async def check_trustline(
        cls, transaction: Transaction, server: Server, locks: Dict
    ):
        """
        Load the destination account json to determine if a trustline has been
        established. If a trustline for the requested asset is found, a the
        transaction is scheduled for processing. If not, the transaction is
        updated to no longer be pending an execution attempt.
        """
        if transaction.to_address.startswith("M"):
            destination_account = MuxedAccount.from_account(
                transaction.to_address
            ).account_id
        else:
            destination_account = transaction.to_address
        try:
            _, account = await get_account_obj_async(
                Keypair.from_public_key(destination_account), server
            )
        except BaseRequestError:
            logger.exception(f"Failed to load account {destination_account}")
            transaction.pending_execution_attempt = False
            await sync_to_async(transaction.save)()
            return
        trustline_found = False
        for balance in account["balances"]:
            if balance.get("asset_type") in ["native", "liquidity_pool_shares"]:
                continue
            if (
                balance["asset_code"] == transaction.asset.code
                and balance["asset_issuer"] == transaction.asset.issuer
            ):
                trustline_found = True
                break
        if trustline_found:
            logger.debug(
                f"detected transaction {transaction.id} is no longer pending trust"
            )
            await cls.process_deposit(transaction, server, locks)
        else:
            transaction.pending_execution_attempt = False
            await sync_to_async(transaction.save)()

    @classmethod
    async def submit(cls, transaction: Transaction, server: Server, locks) -> bool:
        valid_statuses = [
            Transaction.STATUS.pending_user_transfer_start,
            Transaction.STATUS.pending_external,
            Transaction.STATUS.pending_anchor,
            Transaction.STATUS.pending_trust,
        ]
        if transaction.status not in valid_statuses:
            raise ValueError(
                f"Unexpected transaction status: {transaction.status}, expecting "
                f"{' or '.join(valid_statuses)}."
            )

        logger.info(f"initiating Stellar deposit for {transaction.id}")
        transaction.status = Transaction.STATUS.pending_anchor
        await sync_to_async(transaction.save)()
        await maybe_make_callback_async(transaction)

        envelope = None
        if transaction.envelope_xdr:
            try:
                envelope = TransactionEnvelope.from_xdr(
                    transaction.envelope_xdr, settings.STELLAR_NETWORK_PASSPHRASE
                )
            except Exception:
                await sync_to_async(cls.handle_error)(
                    transaction, "Failed to decode transaction envelope"
                )
                await maybe_make_callback_async(transaction)
                return False

        logger.debug(f"requesting lock to submit deposit transaction {transaction.id}")
        async with locks["source_accounts"][transaction.asset.distribution_account]:
            logger.debug(
                f"locked to submit deposit transaction for transaction {transaction.id}"
            )
            if not envelope:
                distribution_acc, _ = await get_account_obj_async(
                    Keypair.from_public_key(transaction.asset.distribution_account),
                    server,
                )
                envelope = await cls.create_deposit_envelope(
                    transaction, distribution_acc, server
                )
                envelope.sign(transaction.asset.distribution_seed)
                transaction.envelope_xdr = envelope.to_xdr()

            transaction.status = Transaction.STATUS.pending_stellar
            await sync_to_async(transaction.save)()
            logger.info(
                f"updating transaction {transaction.id} to pending_stellar status"
            )
            await maybe_make_callback_async(transaction)

            try:
                response = await server.submit_transaction(envelope)
            except (BaseHorizonError, ConnectionError) as e:
                message = getattr(e, "message", str(e))
                await sync_to_async(cls.handle_error)(
                    transaction, f"{e.__class__.__name__}: {message}"
                )
                await maybe_make_callback_async(transaction)
                return False
            logger.debug(f"unlocking after submitting transaction {transaction.id}")

        if not response.get("successful"):
            await sync_to_async(cls.handle_error)(
                transaction,
                "Stellar transaction failed when submitted to horizon: "
                f"{response['result_xdr']}",
            )
            await maybe_make_callback_async(transaction)
            return False
        elif transaction.claimable_balance_supported:
            transaction.claimable_balance_id = cls.get_balance_id(response)

        transaction.paging_token = response["paging_token"]
        transaction.stellar_transaction_id = response["id"]
        transaction.status = Transaction.STATUS.completed
        transaction.completed_at = datetime.datetime.now(datetime.timezone.utc)
        transaction.pending_execution_attempt = False
        if not transaction.quote:
            transaction.amount_out = round(
                Decimal(transaction.amount_in) - Decimal(transaction.amount_fee),
                transaction.asset.significant_decimals,
            )
        await sync_to_async(transaction.save)()
        logger.info(f"transaction {transaction.id} completed.")
        await maybe_make_callback_async(transaction)
        return True

    @classmethod
    async def handle_submit(cls, transaction: Transaction, server, locks):
        logger.debug(f"calling submit() for transaction {transaction.id}")
        try:
            success = await PendingDeposits.submit(transaction, server, locks)
        except Exception as e:
            logger.exception("submit() threw an unexpected exception")
            await sync_to_async(cls.handle_error)(transaction, str(e))
            await maybe_make_callback_async(transaction)
            return

        if success:
            await sync_to_async(transaction.refresh_from_db)()
            try:
                await sync_to_async(rdi.after_deposit)(transaction=transaction)
            except NotImplementedError:
                pass
            except Exception:
                logger.exception("after_deposit() threw an unexpected exception")

    @staticmethod
    async def create_deposit_envelope(
        transaction, source_account, server
    ) -> TransactionEnvelope:
        if transaction.amount_out:
            payment_amount = transaction.amount_out
        elif transaction.quote:
            raise RuntimeError(
                f"transaction {transaction.id} uses a quote but does not have "
                "amount_out assigned"
            )
        else:
            payment_amount = round(
                Decimal(transaction.amount_in) - Decimal(transaction.amount_fee),
                transaction.asset.significant_decimals,
            )
        builder = TransactionBuilder(
            source_account=source_account,
            network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
            # only one operation, so base_fee will be multipled by 1
            base_fee=settings.MAX_TRANSACTION_FEE_STROOPS
            or await server.fetch_base_fee(),
        )
        payment_op_kwargs = {
            "destination": transaction.to_address,
            "asset_code": transaction.asset.code,
            "asset_issuer": transaction.asset.issuer,
            "amount": str(payment_amount),
            "source": transaction.asset.distribution_account,
        }
        if transaction.claimable_balance_supported:
            _, json_resp = await get_account_obj_async(
                Keypair.from_public_key(transaction.to_address), server
            )
            if is_pending_trust(transaction, json_resp):
                claimant = Claimant(destination=transaction.to_address)
                asset = Asset(
                    code=transaction.asset.code, issuer=transaction.asset.issuer
                )
                builder.append_create_claimable_balance_op(
                    claimants=[claimant],
                    asset=asset,
                    amount=str(payment_amount),
                    source=transaction.asset.distribution_account,
                )
            else:
                builder.append_payment_op(**payment_op_kwargs)
        else:
            builder.append_payment_op(**payment_op_kwargs)
        if transaction.memo:
            builder.add_memo(make_memo(transaction.memo, transaction.memo_type))
        return builder.build()

    @staticmethod
    def get_balance_id(response: dict) -> Optional[str]:
        """
        Pulls claimable balance ID from horizon responses if present

        When called we decode and read the result_xdr from the horizon response.
        If any of the operations is a createClaimableBalanceResult we
        decode the Base64 representation of the balanceID xdr.
        After the fact we encode the result to hex.

        The hex representation of the balanceID is important because its the
        representation required to query and claim claimableBalances.

        :param
            response: the response from horizon

        :return
            hex representation of the balanceID
            or
            None (if no createClaimableBalanceResult operation is found)
        """
        result_xdr = response["result_xdr"]
        balance_id_hex = None
        for op_result in TransactionResult.from_xdr(result_xdr).result.results:
            if op_result.tr.type == OperationType.CREATE_CLAIMABLE_BALANCE:
                balance_id_hex = (
                    op_result.tr.create_claimable_balance_result.balance_id.to_xdr_bytes().hex()
                )
        return balance_id_hex

    @classmethod
    async def requires_trustline(
        cls, transaction: Transaction, server: Server, locks: Dict
    ) -> bool:
        try:
            _, pending_trust = await PendingDeposits.get_or_create_destination_account(
                transaction, server, locks
            )
        except RuntimeError as e:
            await sync_to_async(cls.handle_error)(transaction, str(e))
            await maybe_make_callback_async(transaction)
            return True

        if pending_trust and not transaction.claimable_balance_supported:
            logger.info(
                f"destination account is pending_trust for transaction {transaction.id}"
            )
            transaction.status = Transaction.STATUS.pending_trust
            transaction.pending_execution_attempt = False
            await sync_to_async(transaction.save)()
            await maybe_make_callback_async(transaction)
            return True

        return False

    @staticmethod
    async def requires_multisig(transaction: Transaction) -> bool:
        logger.debug(f"checking if transaction {transaction.id} requires multisig")
        master_signer = (
            await transaction.asset.get_distribution_account_master_signer_async()
        )
        thresholds = await transaction.asset.get_distribution_account_thresholds_async()
        return (
            not master_signer
            or master_signer["weight"] == 0
            or master_signer["weight"] < thresholds["med_threshold"]
        )

    @classmethod
    async def save_as_pending_signatures(cls, transaction, server):
        try:
            channel_kp = await sync_to_async(cls.get_channel_keypair)(transaction)
            channel_account, _ = await get_account_obj_async(channel_kp, server)
        except (RuntimeError, ConnectionError) as e:
            transaction.status = Transaction.STATUS.error
            transaction.status_message = str(e)
            logger.error(transaction.status_message)
        else:
            # Create the initial envelope XDR with the channel signature
            envelope = await PendingDeposits.create_deposit_envelope(
                transaction, channel_account, server
            )
            envelope.sign(channel_kp)
            transaction.envelope_xdr = envelope.to_xdr()
            transaction.pending_signatures = True
            transaction.status = Transaction.STATUS.pending_anchor
        transaction.pending_execution_attempt = False
        await sync_to_async(transaction.save)()
        await maybe_make_callback_async(transaction)

    @staticmethod
    def get_channel_keypair(transaction) -> Keypair:
        if not transaction.channel_account:
            logger.info(
                f"calling create_channel_account() for transaction {transaction.id}"
            )
            rdi.create_channel_account(transaction=transaction)
        if not transaction.channel_seed:
            asset = transaction.asset
            quote = transaction.quote
            transaction.refresh_from_db()
            transaction.asset = asset
            transaction.quote = quote
        return Keypair.from_secret(transaction.channel_seed)

    @classmethod
    def handle_error(cls, transaction, message):
        transaction.status_message = message
        transaction.status = Transaction.STATUS.error
        transaction.pending_execution_attempt = False
        transaction.save()
        logger.error(message)



"""Helper functions to use across tests."""
import json
import time
from unittest.mock import Mock

from polaris import settings
from stellar_sdk.keypair import Keypair
from stellar_sdk.transaction_envelope import TransactionEnvelope

TEST_MUXED_ACCOUNT = (
    "MDOP36XYBRBRAFHGRP7VPH335XVEQLSCF2UA7PT34TTZOXFMIGJJWAAAAAAAAAAAPMYT4"
)
TEST_ACCOUNT_MEMO = 123


def mock_check_auth_success(request, func, **kwargs):
    """Mocks `sep10.utils.check_auth`, for success."""
    return func(
        Mock(
            account="test source address",
            muxed_account=None,
            memo=None,
            client_domain=None,
        ),
        request,
        **kwargs,
    )


def mock_check_auth_success_muxed_account(request, func, **kwargs):
    """Mocks `sep10.utils.check_auth`, for success."""

    return func(
        Mock(
            account="GDOP36XYBRBRAFHGRP7VPH335XVEQLSCF2UA7PT34TTZOXFMIGJJXHS5",
            muxed_account=TEST_MUXED_ACCOUNT,
            memo=None,
            client_domain=None,
        ),
        request,
        **kwargs,
    )


def mock_check_auth_success_with_memo(request, func, **kwargs):
    """Mocks `sep10.utils.check_auth`, for success."""

    return func(
        Mock(
            account="test source address",
            muxed_account=None,
            memo=TEST_ACCOUNT_MEMO,
            client_domain=None,
        ),
        request,
        **kwargs,
    )


def mock_check_auth_success_client_domain(request, func, **kwargs):
    """Mocks `sep10.utils.check_auth`, for success."""
    return func(
        Mock(
            account="test source address",
            muxed_account=None,
            memo=None,
            client_domain="test.com",
        ),
        request,
        **kwargs,
    )


def sep10(client, address, seed):
    response = client.get(f"/auth?account={address}", follow=True)
    content = json.loads(response.content)
    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(
        envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE
    )
    client_signing_key = Keypair.from_secret(seed)
    envelope_object.sign(client_signing_key)
    client_signed_envelope_xdr = envelope_object.to_xdr()

    response = client.post(
        "/auth",
        data={"transaction": client_signed_envelope_xdr},
        headers={"Content-Type": "application/json"},
    )
    content = json.loads(response.content)
    encoded_jwt = content["token"]
    assert encoded_jwt
    return encoded_jwt


def interactive_jwt_payload(transaction, transaction_type):
    current_time = time.time()
    return {
        "iss": f"http://testserver/sep24/transactions/{transaction_type}/interactive",
        "exp": current_time + 30,
        "iat": current_time,
        "jti": str(transaction.id),
        "sub": transaction.stellar_account,
    }



from django.urls import path

from polaris.sep38 import info, prices, quote

urlpatterns = [
    path("info", info.info),
    path("prices", prices.get_prices),
    path("price", prices.get_price),
    path("quote", quote.post_quote),
    path("quote/<quote_id>", quote.get_quote),
]



import uuid
from datetime import datetime, timezone, timedelta
from decimal import Decimal

import pytest
import json
from typing import Dict
from unittest.mock import patch, Mock

from stellar_sdk import Keypair, MuxedAccount
from rest_framework.request import Request

from polaris.tests.conftest import USD_DISTRIBUTION_SEED
from polaris.tests.helpers import (
    mock_check_auth_success,
    mock_check_auth_success_client_domain,
    mock_check_auth_success_muxed_account,
    mock_check_auth_success_with_memo,
    TEST_MUXED_ACCOUNT,
    TEST_ACCOUNT_MEMO,
)
from polaris.integrations import WithdrawalIntegration
from polaris.models import Transaction, Asset, OffChainAsset, ExchangePair, Quote
from polaris.sep10.token import SEP10Token
from polaris.utils import make_memo

WITHDRAW_PATH = "/sep6/withdraw"


class GoodWithdrawalIntegration(WithdrawalIntegration):
    def process_sep6_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args,
        **kwargs
    ) -> Dict:
        if params.get("type") == "bad type":
            raise ValueError()
        transaction.save()
        return {"extra_info": {"test": "test"}}


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", GoodWithdrawalIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_good_withdrawal_integration(client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        sep6_enabled=True,
        withdrawal_enabled=True,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        distribution_seed=Keypair.random().secret,
    )
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "type": "bank_account",
            "dest": "test bank account number",
        },
    )
    content = response.json()
    assert response.status_code == 200
    assert content.pop("memo")
    assert content.pop("memo_type") == Transaction.MEMO_TYPES.hash
    assert content == {
        "id": str(Transaction.objects.first().id),
        "account_id": asset.distribution_account,
        "min_amount": round(asset.withdrawal_min_amount, asset.significant_decimals),
        "max_amount": round(asset.withdrawal_max_amount, asset.significant_decimals),
        "extra_info": {"test": "test"},
    }


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", GoodWithdrawalIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success_muxed_account)
def test_success_muxed_account(client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        sep6_enabled=True,
        withdrawal_enabled=True,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        distribution_seed=Keypair.random().secret,
    )
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "type": "bank_account",
            "dest": "test bank account number",
        },
    )
    content = response.json()
    assert response.status_code == 200
    assert content.pop("memo")
    assert content.pop("memo_type") == Transaction.MEMO_TYPES.hash
    assert content == {
        "id": str(Transaction.objects.first().id),
        "account_id": asset.distribution_account,
        "min_amount": round(asset.withdrawal_min_amount, asset.significant_decimals),
        "max_amount": round(asset.withdrawal_max_amount, asset.significant_decimals),
        "extra_info": {"test": "test"},
    }
    assert Transaction.objects.count() == 1
    t = Transaction.objects.first()
    assert t.stellar_account == MuxedAccount.from_account(TEST_MUXED_ACCOUNT).account_id
    assert t.muxed_account == TEST_MUXED_ACCOUNT
    assert t.account_memo is None


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", GoodWithdrawalIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success_with_memo)
def test_success_with_memo(client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        sep6_enabled=True,
        withdrawal_enabled=True,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        distribution_seed=Keypair.random().secret,
    )
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "type": "bank_account",
            "dest": "test bank account number",
        },
    )
    content = response.json()
    assert response.status_code == 200
    assert content.pop("memo")
    assert content.pop("memo_type") == Transaction.MEMO_TYPES.hash
    assert content == {
        "id": str(Transaction.objects.first().id),
        "account_id": asset.distribution_account,
        "min_amount": round(asset.withdrawal_min_amount, asset.significant_decimals),
        "max_amount": round(asset.withdrawal_max_amount, asset.significant_decimals),
        "extra_info": {"test": "test"},
    }
    assert Transaction.objects.count() == 1
    t = Transaction.objects.first()
    assert t.stellar_account == "test source address"
    assert t.muxed_account is None
    assert t.account_memo == TEST_ACCOUNT_MEMO


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_bad_muxed_account(client, acc1_usd_withdrawal_transaction_factory):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    asset = withdraw.asset
    response = client.get(
        WITHDRAW_PATH,
        {"asset_code": asset.code, "type": "good type", "dest": "test", "account": "M"},
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "invalid 'account'"}


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdrawal_success_no_min_max_amounts(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        sep6_enabled=True,
        withdrawal_enabled=True,
        distribution_seed=Keypair.random().secret,
    )
    mock_process_sep6_request.return_value = {
        "extra_info": {"test": "test"},
    }
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "type": "bank_account",
            "dest": "test bank account number",
        },
    )
    mock_process_sep6_request.assert_called_once()
    assert Transaction.objects.count() == 1
    assert response.status_code == 200
    content = response.json()
    assert content.pop("memo")
    assert content.pop("memo_type") == Transaction.MEMO_TYPES.hash
    assert content == {
        "id": str(Transaction.objects.first().id),
        "account_id": asset.distribution_account,
        "extra_info": {"test": "test"},
    }


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdrawal_success_custom_min_max_amounts(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        sep6_enabled=True,
        withdrawal_enabled=True,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        distribution_seed=Keypair.random().secret,
    )
    mock_process_sep6_request.return_value = {
        "extra_info": {"test": "test"},
        "min_amount": 1000,
        "max_amount": 10000,
    }
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "type": "bank_account",
            "dest": "test bank account number",
        },
    )
    mock_process_sep6_request.assert_called_once()
    assert Transaction.objects.count() == 1
    content = response.json()
    assert response.status_code == 200, content
    assert content.pop("memo")
    assert content.pop("memo_type") == Transaction.MEMO_TYPES.hash
    assert content == {
        "id": str(Transaction.objects.first().id),
        "account_id": asset.distribution_account,
        "min_amount": 1000,
        "max_amount": 10000,
        "extra_info": {"test": "test"},
    }


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_bad_memo_type(client, acc1_usd_withdrawal_transaction_factory):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    asset = withdraw.asset
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "type": "good type",
            "dest": "test",
            "memo_type": "none",
        },
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "invalid 'memo_type'"}


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_bad_memo(client, acc1_usd_withdrawal_transaction_factory):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    asset = withdraw.asset
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "type": "good type",
            "dest": "test",
            "memo_type": "id",
            "memo": "not an id",
        },
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "invalid 'memo' for 'memo_type'"}


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", GoodWithdrawalIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_bad_type(client, acc1_usd_withdrawal_transaction_factory):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": withdraw.asset.code,
            "type": "bad type",
            "dest": "test bank account number",
        },
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert "error" in content


class MissingHowDepositIntegration(WithdrawalIntegration):
    def process_sep6_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args,
        **kwargs
    ) -> Dict:
        return {}


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", MissingHowDepositIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_empty_integration_response(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep6])
    response = client.get(
        WITHDRAW_PATH, {"asset_code": asset.code, "type": "good type", "dest": "test"},
    )
    content = json.loads(response.content)
    assert response.status_code == 200
    assert content.pop("memo")
    assert content.pop("memo_type") == Transaction.MEMO_TYPES.hash
    assert content == {
        "id": str(Transaction.objects.first().id),
        "account_id": Keypair.from_secret(USD_DISTRIBUTION_SEED).public_key,
        "min_amount": round(asset.withdrawal_min_amount, asset.significant_decimals),
        "max_amount": round(asset.withdrawal_max_amount, asset.significant_decimals),
        "fee_fixed": round(asset.withdrawal_fee_fixed, asset.significant_decimals),
        "fee_percent": asset.withdrawal_fee_percent,
    }


class BadExtraInfoWithdrawalIntegration(WithdrawalIntegration):
    def process_sep6_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args,
        **kwargs
    ) -> Dict:
        return {"extra_info": "not a dict"}


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", BadExtraInfoWithdrawalIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_bad_extra_info_integration(
    client, acc1_usd_withdrawal_transaction_factory
):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    response = client.get(
        WITHDRAW_PATH,
        {"asset_code": withdraw.asset.code, "type": "good type", "dest": "test"},
    )
    content = json.loads(response.content)
    assert response.status_code == 500
    assert content == {"error": "unable to process the request"}


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_missing_asset(client, acc1_usd_withdrawal_transaction_factory):
    acc1_usd_withdrawal_transaction_factory(protocol=Transaction.PROTOCOL.sep6)
    response = client.get(WITHDRAW_PATH, {"type": "good type", "dest": "test"})
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "invalid 'asset_code' or 'source_asset'"}


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_invalid_asset(client):
    response = client.get(
        WITHDRAW_PATH, {"asset_code": "USD", "type": "good type", "dest": "test"}
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "invalid 'asset_code' or 'source_asset'"}


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_missing_type(client, acc1_usd_withdrawal_transaction_factory):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    response = client.get(
        WITHDRAW_PATH, {"asset_code": withdraw.asset.code, "dest": "test"}
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "'type' is required"}


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_missing_dest(client, acc1_usd_withdrawal_transaction_factory):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    response = client.get(
        WITHDRAW_PATH, {"asset_code": withdraw.asset.code, "type": "good type"}
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "'dest' is required"}


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", GoodWithdrawalIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdrawal_transaction_created(
    client, acc1_usd_withdrawal_transaction_factory
):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    distribution_address = Keypair.from_secret(USD_DISTRIBUTION_SEED).public_key
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": withdraw.asset.code,
            "type": "good type",
            "dest": "test",
            "amount": "100",
        },
    )
    assert response.status_code == 200
    t = (
        Transaction.objects.filter(kind=Transaction.KIND.withdrawal)
        .order_by("-started_at")
        .first()
    )
    assert t
    assert t.memo_type == Transaction.MEMO_TYPES.hash
    assert t.receiving_anchor_account == distribution_address
    assert t.stellar_account == "test source address"
    assert t.amount_in == 100
    assert t.amount_expected == 100
    assert t.asset == withdraw.asset
    assert t.kind == Transaction.KIND.withdrawal
    assert t.status == Transaction.STATUS.pending_user_transfer_start
    assert t.protocol == Transaction.PROTOCOL.sep6


class GoodInfoNeededWithdrawalIntegration(WithdrawalIntegration):
    def process_sep6_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args,
        **kwargs
    ) -> Dict:
        return {
            "type": "non_interactive_customer_info_needed",
            "fields": ["first_name", "last_name"],
        }


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", GoodInfoNeededWithdrawalIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_non_interactive_customer_info_needed(
    client, acc1_usd_withdrawal_transaction_factory
):
    withdraw = acc1_usd_withdrawal_transaction_factory(
        protocol=Transaction.PROTOCOL.sep6
    )
    response = client.get(
        WITHDRAW_PATH,
        {"asset_code": withdraw.asset.code, "type": "good type", "dest": "test"},
    )
    content = json.loads(response.content)
    assert response.status_code == 403
    assert content == {
        "type": "non_interactive_customer_info_needed",
        "fields": ["first_name", "last_name"],
    }


@pytest.mark.django_db
def test_deposit_bad_auth(client):
    response = client.get(WITHDRAW_PATH, {})
    content = json.loads(response.content)
    assert response.status_code == 403
    assert content == {"type": "authentication_required"}


class BadSaveWithdrawalIntegration(WithdrawalIntegration):
    def process_sep6_request(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        transaction: Transaction,
        *args,
        **kwargs
    ) -> Dict:
        transaction.save()
        return {
            "type": "non_interactive_customer_info_needed",
            "fields": ["first_name", "last_name"],
        }


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi", BadSaveWithdrawalIntegration())
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_saved_transaction_on_failure_response(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep6])
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "type": "bank_account",
            "dest": "test bank account number",
        },
    )
    assert response.status_code == 500


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_bad_amount(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep6])
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "account": Keypair.random().public_key,
            "type": "good type",
            "amount": "not an amount",
            "dest": "test bank account number",
        },
    )
    assert response.status_code == 400
    assert "amount" in json.loads(response.content)["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_amount_too_large(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep6])
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "account": Keypair.random().public_key,
            "type": "good type",
            "dest": "test bank account number",
            "amount": asset.deposit_max_amount + 1,
        },
    )
    assert response.status_code == 400
    assert "amount" in json.loads(response.content)["error"]


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_good_amount(mock_deposit, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep6])
    mock_deposit.process_sep6_request = Mock(return_value={"how": "test"})
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "account": Keypair.random().public_key,
            "type": "good type",
            "dest": "test bank account number",
            "amount": asset.deposit_max_amount - 1,
        },
    )
    assert response.status_code == 200
    kwargs = mock_deposit.process_sep6_request.call_args_list[0][1]
    assert kwargs.get("params", {}).get("amount") == asset.deposit_max_amount - 1


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success_client_domain)
def test_withdraw_client_domain_saved(mock_withdraw, client):
    kp = Keypair.random()
    usd = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        sep6_enabled=True,
        withdrawal_enabled=True,
        distribution_seed=Keypair.random().secret,
    )
    mock_withdraw.process_sep6_request = Mock(return_value={"how": "test"})
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": usd.code,
            "account": kp.public_key,
            "type": "good type",
            "dest": "test bank account number",
        },
    )
    content = response.json()
    assert response.status_code == 200, json.dumps(content, indent=2)
    assert Transaction.objects.count() == 1
    transaction = Transaction.objects.first()
    assert transaction.client_domain == "test.com"


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("django.conf.settings.LANGUAGES", [("en", "English"), ("es", "Spansh")])
def test_withdraw_good_lang(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        sep6_enabled=True,
        withdrawal_enabled=True,
    )
    mock_process_sep6_request.return_value = {
        "how": "test",
        "extra_info": {"test": "test"},
    }
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "account": Keypair.random().public_key,
            "lang": "es",
            "type": "good type",
            "dest": "test bank account number",
        },
    )
    assert response.status_code == 200, response.content
    assert Transaction.objects.count() == 1
    mock_process_sep6_request.assert_called_once()


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("django.conf.settings.LANGUAGES", [("en", "English")])
def test_deposit_bad_lang(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        sep6_enabled=True,
        withdrawal_enabled=True,
    )
    mock_process_sep6_request.return_value = {
        "how": "test",
        "extra_info": {"test": "test"},
    }
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "account": Keypair.random().public_key,
            "lang": "es",
            "type": "good type",
            "dest": "test bank account number",
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "unsupported language: es"}


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_bad_on_change_callback(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        sep6_enabled=True,
        withdrawal_enabled=True,
    )
    mock_process_sep6_request.return_value = {
        "how": "test",
        "extra_info": {"test": "test"},
    }
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "account": Keypair.random().public_key,
            "on_change_callback": "invalid domain",
            "type": "good type",
            "dest": "test bank account number",
        },
    )
    mock_process_sep6_request.assert_not_called()
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "invalid callback URL provided"}


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.settings.CALLBACK_REQUEST_DOMAIN_DENYLIST", ["example.com"])
def test_withdraw_denied_on_change_callback(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        sep6_enabled=True,
        withdrawal_enabled=True,
    )
    mock_process_sep6_request.return_value = {
        "how": "test",
        "extra_info": {"test": "test"},
    }
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "account": Keypair.random().public_key,
            "on_change_callback": "https://example.com",
            "type": "good type",
            "dest": "test bank account number",
        },
    )
    mock_process_sep6_request.assert_called_once()
    assert Transaction.objects.count() == 1
    t = Transaction.objects.first()
    assert response.status_code == 200
    assert t.on_change_callback is None


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.settings.CALLBACK_REQUEST_DOMAIN_DENYLIST", ["notexample.com"])
def test_withdraw_good_on_change_callback(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        sep6_enabled=True,
        withdrawal_enabled=True,
    )
    mock_process_sep6_request.return_value = {
        "how": "test",
        "extra_info": {"test": "test"},
    }
    response = client.get(
        WITHDRAW_PATH,
        {
            "asset_code": asset.code,
            "account": Keypair.random().public_key,
            "on_change_callback": "https://example.com",
            "type": "good type",
            "dest": "test bank account number",
        },
    )
    mock_process_sep6_request.assert_called_once()
    assert Transaction.objects.count() == 1
    t = Transaction.objects.first()
    assert response.status_code == 200
    assert t.on_change_callback == "https://example.com"


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_success_indicative_quote(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        sep6_enabled=True,
        withdrawal_enabled=True,
        sep38_enabled=True,
    )
    offchain_asset = OffChainAsset.objects.create(
        scheme="iso4217", identifier="BRL", country_codes="BRA"
    )
    ExchangePair.objects.create(
        buy_asset=offchain_asset.asset_identification_format,
        sell_asset=asset.asset_identification_format,
    )
    mock_process_sep6_request.return_value = {}
    response = client.get(
        WITHDRAW_PATH + "-exchange",
        {
            "destination_asset": offchain_asset.asset_identification_format,
            "source_asset": asset.asset_identification_format,
            "amount": "100.12",
            "type": "good type",
            "dest": "an offchain bank account",
        },
    )
    content = response.json()
    assert response.status_code == 200, content
    assert uuid.UUID(content.pop("id"))
    assert make_memo(content.pop("memo"), "hash")
    assert content == {
        "account_id": asset.distribution_account,
        "min_amount": asset.withdrawal_min_amount,
        "max_amount": asset.withdrawal_max_amount,
        "memo_type": "hash",
    }
    assert Transaction.objects.count() == 1
    t = Transaction.objects.first()
    mock_process_sep6_request.assert_called_once()
    assert t.quote
    assert t.quote.buy_asset == offchain_asset.asset_identification_format
    assert t.quote.sell_asset == asset.asset_identification_format
    assert t.quote.sell_amount == Decimal("100.12")
    assert t.quote.type == Quote.TYPE.indicative


@pytest.mark.django_db
@patch("polaris.sep6.withdraw.rwi.process_sep6_request")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_withdraw_success_firm_quote(mock_process_sep6_request, client):
    asset = Asset.objects.create(
        code="USD",
        issuer=Keypair.random().public_key,
        withdrawal_min_amount=10,
        withdrawal_max_amount=1000,
        sep6_enabled=True,
        withdrawal_enabled=True,
        sep38_enabled=True,
        distribution_seed=Keypair.random().secret,
    )
    offchain_asset = OffChainAsset.objects.create(
        scheme="iso4217", identifier="BRL", country_codes="BRA"
    )
    ExchangePair.objects.create(
        buy_asset=offchain_asset.asset_identification_format,
        sell_asset=asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=uuid.uuid4(),
        stellar_account="test source address",
        buy_asset=offchain_asset.asset_identification_format,
        sell_asset=asset.asset_identification_format,
        price=Decimal(1),
        sell_amount=Decimal("102.12"),
        buy_amount=Decimal("102.12"),
        type=Quote.TYPE.firm,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    mock_process_sep6_request.return_value = {}
    response = client.get(
        WITHDRAW_PATH + "-exchange",
        {
            "source_asset": asset.asset_identification_format,
            "destination_asset": offchain_asset.asset_identification_format,
            "quote_id": str(quote.id),
            "amount": "102.12",
            "type": "good type",
            "dest": "test bank account number",
        },
    )
    content = response.json()
    assert response.status_code == 200, content
    assert uuid.UUID(content.pop("id"))
    assert make_memo(content.pop("memo"), "hash")
    assert content == {
        "account_id": asset.distribution_account,
        "min_amount": asset.withdrawal_min_amount,
        "max_amount": asset.withdrawal_max_amount,
        "memo_type": "hash",
    }
    assert Transaction.objects.count() == 1
    t = Transaction.objects.first()
    mock_process_sep6_request.assert_called_once()
    assert t.quote.id == quote.id
    assert t.quote.type == Quote.TYPE.firm



import pytest
from unittest.mock import patch, Mock
from secrets import token_bytes
from requests.exceptions import RequestException

from stellar_sdk import Keypair, TextMemo, IdMemo, HashMemo
from stellar_sdk.exceptions import NotFoundError
from stellar_sdk.account import Thresholds

from polaris import utils
from polaris import settings
from polaris.models import Transaction

test_module = "polaris.utils"


def test_load_account():
    mock_response = {
        "sequence": 1,
        "account_id": Keypair.random().public_key,
        "signers": [
            {
                "key": Keypair.random().public_key,
                "weight": 1,
                "type": "ed25519_public_key",
            }
        ],
        "thresholds": {"low_threshold": 0, "med_threshold": 1, "high_threshold": 2},
    }
    account = utils.load_account(mock_response)
    assert account.account.account_id == mock_response["account_id"]
    assert account.sequence == mock_response["sequence"]

    account.load_ed25519_public_key_signers()
    assert account.signers == mock_response["signers"]
    assert account.thresholds == Thresholds(
        mock_response["thresholds"]["low_threshold"],
        mock_response["thresholds"]["med_threshold"],
        mock_response["thresholds"]["high_threshold"],
    )


@patch(f"{test_module}.settings.HORIZON_SERVER.accounts")
@patch(f"{test_module}.load_account")
def test_get_account_obj(mock_load_account, mock_accounts_endpoint):
    kp = Keypair.random()
    mock_accounts_endpoint.return_value.account_id.return_value.call.return_value = {
        "sequence": 1,
        "account_id": kp.public_key,
        "signers": [{"key": kp.public_key, "weight": 1, "type": "ed25519_public_key"}],
        "thresholds": {"low_threshold": 0, "med_threshold": 1, "high_threshold": 2},
    }
    account, mock_json = utils.get_account_obj(kp)
    mock_accounts_endpoint.return_value.account_id.assert_called_once_with(
        account_id=kp.public_key
    )
    mock_accounts_endpoint.return_value.account_id.return_value.call.assert_called_once()
    assert (
        mock_json
        == mock_accounts_endpoint.return_value.account_id.return_value.call.return_value
    )
    mock_load_account.assert_called_once_with(mock_json)


@patch(f"{test_module}.settings.HORIZON_SERVER.accounts")
@patch(f"{test_module}.load_account")
def test_get_account_obj_not_found(mock_load_account, mock_accounts_endpoint):
    mock_accounts_endpoint.return_value.account_id.return_value.call.side_effect = NotFoundError(
        Mock()
    )
    kp = Keypair.random()
    with pytest.raises(RuntimeError, match=f"account {kp.public_key} does not exist"):
        utils.get_account_obj(kp)
    mock_load_account.assert_not_called()


def test_memo_str_none():
    assert utils.memo_str(None) == (None, None)


def test_memo_str_bad_input_type():
    with pytest.raises(ValueError):
        utils.memo_str(Mock())


def test_memo_str_text_memo():
    assert utils.memo_str(TextMemo("test")) == ("test", Transaction.MEMO_TYPES.text)


def test_memo_str_id_memo():
    assert utils.memo_str(IdMemo(123)) == ("123", Transaction.MEMO_TYPES.id)


def test_memo_str_hash_memo():
    raw_bytes = token_bytes(32)
    memo_str = utils.memo_hex_to_base64(raw_bytes.hex())
    assert utils.memo_str(HashMemo(raw_bytes)) == (
        memo_str,
        Transaction.MEMO_TYPES.hash,
    )


@patch(f"{test_module}.post")
@patch(f"{test_module}.TransactionSerializer")
def test_make_on_change_callback_success(mock_serializer, mock_post):
    mock_transaction = Mock(on_change_callback="test")
    utils.make_on_change_callback(mock_transaction)
    mock_serializer.assert_called_once_with(mock_transaction)
    mock_post.assert_called_once_with(
        url=mock_transaction.on_change_callback,
        json=mock_serializer(mock_transaction).data,
        timeout=settings.CALLBACK_REQUEST_TIMEOUT,
    )


@patch(f"{test_module}.post")
@patch(f"{test_module}.TransactionSerializer")
def test_make_on_change_callback_success_with_timeout(mock_serializer, mock_post):
    mock_transaction = Mock(on_change_callback="test")
    utils.make_on_change_callback(mock_transaction, timeout=5)
    mock_serializer.assert_called_once_with(mock_transaction)
    mock_post.assert_called_once_with(
        url=mock_transaction.on_change_callback,
        json=mock_serializer(mock_transaction).data,
        timeout=5,
    )


@patch(f"{test_module}.post")
@patch(f"{test_module}.TransactionSerializer")
def test_make_on_change_callback_raises_valueerror_for_postmessage(
    mock_serializer, mock_post
):
    mock_transaction = Mock(on_change_callback="postMessage")
    with pytest.raises(ValueError, match="invalid or missing on_change_callback"):
        utils.make_on_change_callback(mock_transaction)
    mock_serializer.assert_not_called()
    mock_post.assert_not_called()


@patch(f"{test_module}.make_on_change_callback")
@patch(f"{test_module}.logger.error")
def test_maybe_make_callback_success(mock_log_error, mock_make_callback):
    mock_transaction = Mock()
    utils.make_on_change_callback(mock_transaction)
    mock_make_callback.assert_called_once_with(mock_transaction)
    mock_log_error.assert_not_called()


@patch(f"{test_module}.make_on_change_callback")
@patch(f"{test_module}.logger.error")
def test_maybe_make_callback_not_ok(mock_log_error, mock_make_callback):
    mock_make_callback.return_value = Mock(ok=False)
    mock_transaction = Mock()
    utils.maybe_make_callback(mock_transaction)
    mock_make_callback.assert_called_once_with(mock_transaction, timeout=None)
    mock_log_error.assert_called_once()


@patch(f"{test_module}.make_on_change_callback")
@patch(f"{test_module}.logger.error")
def test_maybe_make_callback_raises(mock_log_error, mock_make_callback):
    mock_make_callback.side_effect = RequestException()
    mock_transaction = Mock()
    utils.maybe_make_callback(mock_transaction)
    mock_make_callback.assert_called_once_with(mock_transaction, timeout=None)
    mock_log_error.assert_called_once()


@patch(f"{test_module}.make_on_change_callback")
@patch(f"{test_module}.logger.error")
def test_maybe_make_callback_postmessage(mock_log_error, mock_make_callback):
    mock_transaction = Mock(on_change_callback="postMessage")
    utils.maybe_make_callback(mock_transaction)
    mock_make_callback.assert_not_called()
    mock_log_error.assert_not_called()



import pytest

from stellar_sdk import Keypair

from polaris.models import Asset, OffChainAsset, DeliveryMethod

ENDPOINT = "/sep38/info"


@pytest.mark.django_db
def test_info(client):
    usd_stellar = Asset.objects.create(
        code="usd", issuer=Keypair.random().public_key, sep38_enabled=True
    )
    brl_offchain = OffChainAsset.objects.create(
        scheme="iso4217", identifier="BRL", country_codes="BRA"
    )
    delivery_methods = [
        DeliveryMethod.objects.create(
            type=DeliveryMethod.TYPE.buy, name="cash_pickup", description="cash pick-up"
        ),
        DeliveryMethod.objects.create(
            type=DeliveryMethod.TYPE.sell,
            name="cash_dropoff",
            description="cash drop-off",
        ),
    ]
    brl_offchain.delivery_methods.add(*delivery_methods)

    response = client.get(ENDPOINT)
    assert response.status_code == 200, response.content

    body = response.json()
    assert len(body["assets"]) == 2, body

    expected_usd_stellar = {"asset": f"stellar:{usd_stellar.code}:{usd_stellar.issuer}"}
    expected_brl_offchain = {
        "asset": brl_offchain.asset_identification_format,
        "sell_delivery_methods": [
            {"name": "cash_dropoff", "description": "cash drop-off"}
        ],
        "buy_delivery_methods": [
            {"name": "cash_pickup", "description": "cash pick-up"}
        ],
        "country_codes": ["BRA"],
    }
    for a in body["assets"]:
        if a["asset"] == expected_usd_stellar["asset"]:
            assert a == expected_usd_stellar, (a, expected_usd_stellar)
        elif a["asset"] == brl_offchain.asset_identification_format:
            assert a == expected_brl_offchain, (a, expected_brl_offchain)
        else:
            raise ValueError(f"unexpected asset: {a}")



import pytest
from decimal import Decimal
from unittest.mock import patch, Mock

from stellar_sdk import Keypair

from polaris.models import Asset, OffChainAsset, DeliveryMethod, ExchangePair
from polaris.tests.helpers import mock_check_auth_success


PRICES_ENDPOINT = "/sep38/prices"
code_path = "polaris.sep38.prices"


def default_data():
    usd_stellar = Asset.objects.create(
        code="usd", issuer=Keypair.random().public_key, sep38_enabled=True
    )
    brl_offchain = OffChainAsset.objects.create(
        scheme="iso4217", identifier="BRL", country_codes="BRA"
    )
    delivery_methods = [
        DeliveryMethod.objects.create(
            type=DeliveryMethod.TYPE.buy,
            name="cash_pickup",
            description="cash pick-up",
        ),
        DeliveryMethod.objects.create(
            type=DeliveryMethod.TYPE.sell,
            name="cash_dropoff",
            description="cash drop-off",
        ),
    ]
    brl_offchain.delivery_methods.add(*delivery_methods)
    pair = ExchangePair.objects.create(
        buy_asset=brl_offchain.asset_identification_format,
        sell_asset=usd_stellar.asset_identification_format,
    )
    return {
        "stellar_assets": [usd_stellar],
        "offchain_assets": [brl_offchain],
        "exchange_pairs": [pair],
        "delivery_methods": delivery_methods,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_success_no_optional_params(mock_rqi, client):
    data = default_data()
    mock_rqi.get_prices = Mock(return_value=[Decimal(2.123)])
    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 200, response.content
    body = response.json()
    assert body == {
        "buy_assets": [
            {
                "asset": data["offchain_assets"][0].asset_identification_format,
                "price": "2.12",
                "decimals": data["offchain_assets"][0].significant_decimals,
            }
        ]
    }
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["stellar_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["offchain_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": None,
        "country_code": None,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_success_country_code_buy_delivery_method(mock_rqi, client):
    data = default_data()
    mock_rqi.get_prices = Mock(return_value=[Decimal(2.123)])
    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "country_code": "BRA",
            "buy_delivery_method": "cash_pickup",
        },
    )
    assert response.status_code == 200, response.content
    body = response.json()
    assert body == {
        "buy_assets": [
            {
                "asset": data["offchain_assets"][0].asset_identification_format,
                "price": "2.12",
                "decimals": data["offchain_assets"][0].significant_decimals,
            }
        ]
    }
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["stellar_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["offchain_assets"],
        "buy_delivery_method": data["delivery_methods"][0],
        "sell_delivery_method": None,
        "country_code": "BRA",
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_success_country_code_sell_delivery_method(mock_rqi, client):
    data = default_data()
    # swap exchange pair
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()

    mock_rqi.get_prices = Mock(return_value=[Decimal(2.123)])
    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "country_code": "BRA",
            "sell_delivery_method": "cash_dropoff",
        },
    )
    assert response.status_code == 200, response.content
    body = response.json()
    assert body == {
        "buy_assets": [
            {
                "asset": data["stellar_assets"][0].asset_identification_format,
                "price": "2.12",
                "decimals": data["stellar_assets"][0].significant_decimals,
            }
        ]
    }
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["offchain_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["stellar_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": data["delivery_methods"][1],
        "country_code": "BRA",
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_success_no_optional_params_swap_exchange_pair(mock_rqi, client):
    data = default_data()
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()

    mock_rqi.get_prices = Mock(return_value=[Decimal(2.123)])
    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 200, response.content
    body = response.json()
    assert body == {
        "buy_assets": [
            {
                "asset": data["stellar_assets"][0].asset_identification_format,
                "price": "2.12",
                "decimals": data["stellar_assets"][0].significant_decimals,
            }
        ]
    }
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["offchain_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["stellar_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": None,
        "country_code": None,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_success_no_exchange_pairs(mock_rqi, client):
    data = default_data()
    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'buy_assets' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_missing_sell_amount(mock_rqi, client):
    data = default_data()
    response = client.get(
        PRICES_ENDPOINT,
        {"sell_asset": data["stellar_assets"][0].asset_identification_format},
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "missing required parameters. Required: sell_asset, sell_amount"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_missing_sell_asset(mock_rqi, client):
    default_data()
    response = client.get(PRICES_ENDPOINT, {"sell_amount": 100})
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "missing required parameters. Required: sell_asset, sell_amount"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_both_delivery_methods(mock_rqi, client):
    data = default_data()
    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_delivery_method": "cash_pickup",
            "sell_delivery_method": "cash_dropoff",
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "'buy_delivery_method' or 'sell_delivery_method' is valid, but not both"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_sell_stellar_with_sell_delivery_method(mock_rqi, client):
    data = default_data()
    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "sell_delivery_method": "cash_dropoff",
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "unexpected 'sell_delivery_method', client intends to sell a Stellar asset"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_sell_offchain_with_buy_delivery_method(mock_rqi, client):
    data = default_data()

    # swap exchange pair
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_delivery_method": "cash_pickup",
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "unexpected 'buy_delivery_method', client intends to buy a Stellar asset"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_bad_stellar_format(mock_rqi, client):
    default_data()
    response = client.get(
        PRICES_ENDPOINT, {"sell_asset": f"stellar:USDC", "sell_amount": 100,}
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "invalid 'sell_asset' format"}
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_bad_offchain_format(mock_rqi, client):
    default_data()
    response = client.get(PRICES_ENDPOINT, {"sell_asset": f"USD", "sell_amount": 100,})
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "invalid 'sell_asset' format"}
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_stellar_asset_not_found(mock_rqi, client):
    data = default_data()

    # delete stellar asset from DB
    data["stellar_assets"][0].delete()

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'sell_asset' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_offchain_asset_not_found(mock_rqi, client):
    data = default_data()

    # swap exchange pair
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()
    # delete offchain asset from DB
    data["offchain_assets"][0].delete()

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'sell_asset' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_bad_buy_delivery_method(mock_rqi, client):
    data = default_data()

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_delivery_method": "bad_delivery_method",
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'buy_assets' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_bad_country_code(mock_rqi, client):
    data = default_data()

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "country_code": "TEST",
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'buy_assets' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_bad_sell_delivery_method(mock_rqi, client):
    data = default_data()
    # swap exchange pair
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "sell_delivery_method": "bad_delivery_method",
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'sell_asset' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_bad_sell_amount(mock_rqi, client):
    data = default_data()

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": "test",
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "invalid 'sell_amount'; Expected decimal string."
    }
    mock_rqi.get_prices.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_not_enough_prices_returned(mock_rqi, client):
    data = default_data()
    mock_rqi.get_prices.return_value = []

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 500, response.content
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["stellar_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["offchain_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": None,
        "country_code": None,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_bad_return_type(mock_rqi, client):
    data = default_data()
    mock_rqi.get_prices.return_value = None

    with pytest.raises(Exception):
        client.get(
            PRICES_ENDPOINT,
            {
                "sell_asset": data["stellar_assets"][0].asset_identification_format,
                "sell_amount": 100,
            },
        )
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["stellar_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["offchain_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": None,
        "country_code": None,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_num_prices_dont_match(mock_rqi, client):
    data = default_data()
    mock_rqi.get_prices.return_value = [1, 2]

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 500, response.content
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["stellar_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["offchain_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": None,
        "country_code": None,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_anchor_raises_value_error(mock_rqi, client):
    data = default_data()
    mock_rqi.get_prices.side_effect = ValueError("test")

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "test"}
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["stellar_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["offchain_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": None,
        "country_code": None,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_anchor_raises_runtime_error(mock_rqi, client):
    data = default_data()
    mock_rqi.get_prices.side_effect = RuntimeError("test")

    response = client.get(
        PRICES_ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
        },
    )
    assert response.status_code == 503, response.content
    assert response.json() == {"error": "test"}
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["stellar_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["offchain_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": None,
        "country_code": None,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_prices_failure_anchor_raises_unexpected_error(mock_rqi, client):
    data = default_data()
    mock_rqi.get_prices.side_effect = IndexError("test")

    with pytest.raises(Exception):
        client.get(
            PRICES_ENDPOINT,
            {
                "sell_asset": data["stellar_assets"][0].asset_identification_format,
                "sell_amount": 100,
            },
        )
    mock_rqi.get_prices.assert_called_once()
    kwargs = mock_rqi.get_prices.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert kwargs == {
        "sell_asset": data["stellar_assets"][0],
        "sell_amount": Decimal(100),
        "buy_assets": data["offchain_assets"],
        "buy_delivery_method": None,
        "sell_delivery_method": None,
        "country_code": None,
    }



import json
import uuid
from unittest.mock import Mock, patch

import pytest
from polaris.tests.helpers import mock_check_auth_success
from polaris.models import Transaction, Asset


endpoint = "/sep31/transactions/"


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_success_update(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep31,
        kind=Transaction.KIND.send,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"fields": {"transaction": {"bank_account": "foo"}}},
        content_type="application/json",
    )
    transaction.refresh_from_db()
    assert response.status_code == 200
    assert transaction.required_info_updates is None
    assert transaction.required_info_message is None
    assert transaction.status == Transaction.STATUS.pending_receiver


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_bad_id(client):
    response = client.patch(
        endpoint + "not an id",
        {"fields": {"transaction": {"bank_account": "foo"}}},
        content_type="application/json",
    )
    assert response.status_code == 404


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_bad_id(mock_sep31_integration, client):
    mock_sep31_integration.valid_sending_anchor.return_value = False
    response = client.patch(
        endpoint + "test",
        {"fields": {"transaction": {"bank_account": "foo"}}},
        content_type="application/json",
    )
    assert response.status_code == 403
    assert mock_sep31_integration.valid_sending_anchor.was_called_once_with(
        "test source address"
    )


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_missing_id(client):
    response = client.patch(
        endpoint,
        {"fields": {"transaction": {"bank_account": "foo"}}},
        content_type="application/json",
    )
    assert response.status_code == 404


def test_no_auth(client):
    response = client.patch(endpoint + "test")
    assert response.status_code == 403


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_not_found(client):
    response = client.patch(
        endpoint + str(uuid.uuid4()),
        {"fields": {"transaction": {"bank_account": "foo"}}},
        content_type="application/json",
    )
    assert response.status_code == 404


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_update_not_required(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep31,
        kind=Transaction.KIND.send,
        status=Transaction.STATUS.pending_sender,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"fields": {"transaction": {"bank_account": "foo"}}},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert "update not required" in response.json()["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_bad_info_update_column(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep31,
        kind=Transaction.KIND.send,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates="test",
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"fields": {"transaction": {"bank_account": "foo"}}},
        content_type="application/json",
    )
    assert response.status_code == 500


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_bad_update_body(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep31,
        kind=Transaction.KIND.send,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"fields": {"transaction": {"not a listed field": True}}},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert "missing bank_account" in response.json()["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_missing_update_category(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep31,
        kind=Transaction.KIND.send,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"fields": {"first_name": "test"}},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert "missing transaction" in response.json()["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_bad_category_value_type(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep31,
        kind=Transaction.KIND.send,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"fields": {"transaction": "not a dict"}},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert "invalid type" in response.json()["error"]


raise_error_integration = Mock(
    process_patch_request=Mock(side_effect=ValueError("test"))
)


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    raise_error_integration,
)
def test_user_defined_exception(client):
    asset = Asset.objects.create(code="USD")
    transaction = Transaction.objects.create(
        asset=asset,
        stellar_account="test source address",
        protocol=Transaction.PROTOCOL.sep31,
        kind=Transaction.KIND.send,
        status=Transaction.STATUS.pending_transaction_info_update,
        required_info_message="The provided 'bank_account' was invalid.",
        required_info_updates=json.dumps(
            {"transaction": {"bank_account": {"description": "a description"}}}
        ),
    )
    response = client.patch(
        endpoint + str(transaction.id),
        {"fields": {"transaction": {"bank_account": "foo"}}},
        content_type="application/json",
    )
    assert response.status_code == 400
    assert response.json()["error"] == "test"



import urllib3
from decimal import Decimal
from typing import Optional, Dict
from logging import getLogger

from django.core.management.base import BaseCommand, CommandError
from django.db.models import Q
from stellar_sdk import Keypair, TransactionBuilder, Server
from stellar_sdk.account import Account, Thresholds
from stellar_sdk.exceptions import (
    NotFoundError,
    BaseHorizonError,
    Ed25519SecretSeedInvalidError,
)

from polaris import settings
from polaris.models import Transaction, Asset

logger = getLogger(__name__)


class Command(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset_parser = None
        self.issue_parser = None
        self.server = Server("https://horizon-testnet.stellar.org")
        self.http = urllib3.PoolManager()

    def add_arguments(self, parser):  # pragma: no cover
        subparsers = parser.add_subparsers(dest="subcommands", required=True)
        self.reset_parser = subparsers.add_parser(
            "reset", help="a sub-command for testnet resets"
        )
        self.issue_parser = subparsers.add_parser(
            "issue", help="a sub-command for issuing assets on testnet"
        )
        self.issue_parser.add_argument(
            "--asset", "-a", help="the code of the asset issued"
        )
        self.issue_parser.add_argument(
            "--issuer-seed", "-i", help="the issuer's secret key"
        )
        self.issue_parser.add_argument(
            "--distribution-seed", "-d", help="the distribution account's secret key"
        )
        self.issue_parser.add_argument(
            "--client-seed", "-c", help="the client account's secret key"
        )
        self.issue_parser.add_argument(
            "--issue-amount",
            type=Decimal,
            help="the amount sent to distribution account. Also the limit for the trustline.",
        )
        self.issue_parser.add_argument(
            "--client-amount",
            type=Decimal,
            help="the amount sent to client account. Also the limit for the trustline.",
        )

    def handle(self, *_args, **options):  # pragma: no cover
        if options.get("subcommands") == "reset":
            self.reset(**options)
        elif options.get("subcommands") == "issue":
            self.issue(**options)

    def reset(self, **options):
        """
        Perform any necessary functions to ensure the anchor is in a valid
        state after a testnet reset. Currently this involves the following:

        - re-issuing every Asset object in the DB
        - moves all `pending_trust` Transactions to `error`
        - setting the most-recently streamed Transaction object's
            `paging_token` attribute to None. This signals to
            watch_transactions to stream using the `"now"` keyword.
        """
        print("\nResetting each asset's most recent paging token")
        for asset in Asset.objects.filter(distribution_seed__isnull=False):
            transaction = (
                Transaction.objects.filter(
                    Q(kind=Transaction.KIND.withdrawal) | Q(kind=Transaction.KIND.send),
                    receiving_anchor_account=asset.distribution_account,
                    status=Transaction.STATUS.completed,
                )
                .order_by("-completed_at")
                .first()
            )
            if transaction:
                transaction.paging_token = None
                transaction.save()
        print("\nPlacing all pending_trust transactions into error status")
        Transaction.objects.filter(status=Transaction.STATUS.pending_trust).update(
            status=Transaction.STATUS.error
        )
        for asset in Asset.objects.filter(issuer__isnull=False):
            print(f"\nIssuing {asset.code}")
            issuer_seed = input(f"Seed for {asset.code} issuer (enter to skip): ")
            if not issuer_seed:
                continue
            try:
                Keypair.from_secret(issuer_seed)
            except Ed25519SecretSeedInvalidError:
                raise CommandError("Bad seed string for issuer account")
            distribution_seed = asset.distribution_seed
            if not distribution_seed:
                distribution_seed = input(
                    f"Seed for {asset.code} distribution account: "
                )
                try:
                    Keypair.from_secret(distribution_seed)
                except Ed25519SecretSeedInvalidError:
                    raise CommandError("Bad seed string for distribution account")
            self.issue(
                **{
                    "asset": asset.code,
                    "issuer_seed": issuer_seed,
                    "distribution_seed": distribution_seed,
                    "issue_amount": Decimal(10000000),
                }
            )

    def issue(self, **options):
        """
        Issue the asset specified using the `options` passed to the subcommand
        or function call. The code here is a port of the following project:

        https://github.com/stellar/create-stellar-token

        Users can setup distribution and client accounts for the asset being
        issued as well.
        """
        code = options.get("asset") or "TEST"
        issuer = Keypair.from_secret(
            options.get("issuer_seed") or Keypair.random().secret
        )
        distributor = Keypair.from_secret(
            options.get("distribution_seed") or Keypair.random().secret
        )
        client, client_amount = None, None
        if options.get("client_seed") or options.get("client_amount"):
            client = Keypair.from_secret(
                options.get("client_seed") or Keypair.random().secret
            )
            client_amount = options.get("client_amount") or Decimal(1000)
        issue_amount = options.get("issue_amount") or Decimal(100000)

        print("\nIssuer account public and private keys:")
        print(f"public key: {issuer.public_key}")
        print(f"secret key: {issuer.secret}\n")
        print("Distribution account public and private keys:")
        print(f"public key: {distributor.public_key}")
        print(f"secret key: {distributor.secret}\n")
        print(f"Issuing {issue_amount} {code} to the distribution account")
        if client:
            print("\nClient account public and private keys:")
            print(f"public key: {client.public_key}")
            print(f"secret key: {client.secret}\n")
            print(f"Sending {client_amount} to the client account\n")

        accounts = {}
        for kp in [issuer, distributor, client]:
            if not kp:
                continue
            try:
                accounts[kp.public_key] = (
                    self.server.accounts().account_id(kp.public_key).call()
                )
            except NotFoundError:
                print(f"Funding {kp.public_key} ...")
                self.http.request(
                    "GET", f"https://friendbot.stellar.org?addr={kp.public_key}"
                )
                accounts[kp.public_key] = (
                    self.server.accounts().account_id(kp.public_key).call()
                )

        self.add_balance(code, issue_amount, accounts, distributor, issuer, issuer)
        if client:
            self.add_balance(code, client_amount, accounts, client, distributor, issuer)

        home_domain = input("Home domain for the issuing account (enter to skip): ")
        if home_domain:
            self.set_home_domain(issuer, home_domain)

    def set_home_domain(self, issuer: Keypair, home_domain: str):
        envelope = (
            TransactionBuilder(
                self.server.load_account(issuer.public_key),
                base_fee=settings.MAX_TRANSACTION_FEE_STROOPS
                or self.server.fetch_base_fee(),
                network_passphrase="Test SDF Network ; September 2015",
            )
            .append_set_options_op(home_domain=home_domain)
            .set_timeout(30)
            .build()
        )
        envelope.sign(issuer)
        try:
            self.server.submit_transaction(envelope)
        except BaseHorizonError as e:
            print(
                f"Failed to set {home_domain} as home_domain for {issuer.public_key}."
                f"Result codes: {e.extras.get('result_codes')}"
            )
        else:
            print("Success!")

    def add_balance(self, code, amount, accounts, dest, src, issuer):
        tb = TransactionBuilder(
            self.server.load_account(src.public_key),
            base_fee=settings.MAX_TRANSACTION_FEE_STROOPS
            or self.server.fetch_base_fee(),
            network_passphrase="Test SDF Network ; September 2015",
        )
        balance = self.get_balance(code, issuer.public_key, accounts[dest.public_key])
        if not balance:
            print(f"\nCreating {code} trustline for {dest.public_key}")
            if settings.MAX_TRANSACTION_FEE_STROOPS:
                # halve the base_fee because there are 2 operations
                tb.base_fee = tb.base_fee // 2
            tb.append_change_trust_op(
                asset_code=code, asset_issuer=issuer.public_key, source=dest.public_key
            )
            payment_amount = amount
        elif Decimal(balance) < amount:
            print(f"\nReplenishing {code} balance to {amount} for {dest.public_key}")
            payment_amount = amount - Decimal(balance)
        else:
            print(
                "Destination account already has more than the amount "
                "specified, skipping\n"
            )
            return

        print(f"Sending {code} payment of {payment_amount} to {dest.public_key}")
        tb.append_payment_op(
            destination=dest.public_key,
            amount=payment_amount,
            asset_code=code,
            asset_issuer=issuer.public_key,
            source=src.public_key,
        )
        envelope = tb.set_timeout(30).build()
        if len(tb.operations) == 2:
            # add destination's signature if we're adding a trustline
            envelope.sign(dest)
        envelope.sign(src)
        try:
            self.server.submit_transaction(envelope)
        except BaseHorizonError as e:
            print(
                f"Failed to send {code} payment to {dest.public_key}. "
                f"Result codes: {e.extras.get('result_codes')}"
            )
        else:
            print("Success!")

    def account_from_json(self, json):
        sequence = int(json["sequence"])
        thresholds = Thresholds(
            json["thresholds"]["low_threshold"],
            json["thresholds"]["med_threshold"],
            json["thresholds"]["high_threshold"],
        )
        account = Account(account_id=json["id"], sequence=sequence)
        account.signers = json["signers"]
        account.thresholds = thresholds
        return account

    def get_balance(self, code, issuer_public_key, json) -> Optional[str]:
        for balance_obj in json["balances"]:
            if (
                balance_obj.get("asset_code") == code
                and balance_obj.get("asset_issuer") == issuer_public_key
            ):
                return balance_obj["balance"]



import pytest
from datetime import datetime
from unittest.mock import Mock, patch

from polaris.models import Transaction
from polaris.management.commands.poll_outgoing_transactions import Command


mock_return_passed_transactions = Mock(poll_outgoing_transactions=lambda x: list(x))


@pytest.mark.django_db
@patch(
    "polaris.management.commands.poll_outgoing_transactions.rri",
    mock_return_passed_transactions,
)
def test_ready_transaction(acc1_usd_deposit_transaction_factory):
    transaction = acc1_usd_deposit_transaction_factory(
        protocol=Transaction.PROTOCOL.sep31
    )
    transaction.status = Transaction.STATUS.pending_external
    transaction.save()
    Command.poll_outgoing_transactions()
    transaction.refresh_from_db()
    assert transaction.status == Transaction.STATUS.completed
    assert isinstance(transaction.completed_at, datetime)


@pytest.mark.django_db
@patch(
    "polaris.management.commands.poll_outgoing_transactions.rri",
    Mock(poll_outgoing_transactions=Mock(side_effect=ValueError())),
)
def test_error_raised_in_integration(acc1_usd_deposit_transaction_factory):
    transaction = acc1_usd_deposit_transaction_factory(
        protocol=Transaction.PROTOCOL.sep31
    )
    transaction.status = Transaction.STATUS.pending_external
    transaction.save()
    Command.poll_outgoing_transactions()
    assert transaction.status == Transaction.STATUS.pending_external
    assert not transaction.completed_at


@pytest.mark.django_db
@patch(
    "polaris.management.commands.poll_outgoing_transactions.rri",
    Mock(poll_outgoing_transactions=Mock(return_value=[1, 2, 3])),
)
def test_bad_integration_return_value(acc1_usd_deposit_transaction_factory):
    transaction = acc1_usd_deposit_transaction_factory(
        protocol=Transaction.PROTOCOL.sep31
    )
    transaction.status = Transaction.STATUS.pending_external
    transaction.save()
    Command.poll_outgoing_transactions()
    transaction.refresh_from_db()
    assert transaction.status == Transaction.STATUS.pending_external
    assert not transaction.completed_at



"""This module tests the `/info` endpoint."""
import json
import pytest


def _get_expected_response():
    return """
        {
            "deposit": {
                "USD": {
                    "enabled": true,
                    "min_amount": 0.1,
                    "max_amount": 1000.0,
                    "fee_fixed": 5.0,
                    "fee_percent": 1.0
                },
                "ETH": {
                    "enabled": true,
                    "max_amount": 10000000.0,
                    "fee_fixed": 0.002,
                    "fee_percent": 0.0
                }
            },
            "withdraw": {
                "USD": {
                    "enabled": true,
                    "min_amount": 0.1,
                    "max_amount": 1000.0,
                    "fee_fixed": 5.0,
                    "fee_percent": 0.0
                },
                "ETH": {
                    "enabled": false
                }
            },
            "fee": {
                "enabled": true
            },
            "features": {
                "account_creation": true,
                "claimable_balances": true
            }
        }
    """


@pytest.mark.django_db
def test_info_endpoint(client, usd_asset_factory, eth_asset_factory):
    """
    Ensures the /info endpoint provides data in the expected format, according to
    SEP 24: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md
    """
    usd_asset_factory()
    eth_asset_factory()

    response = client.get(f"/sep24/info", follow=True)
    content = json.loads(response.content)
    expected_response = json.loads(_get_expected_response())
    assert content == expected_response
    assert response.status_code == 200



"""This module defines custom management commands for the app admin."""
import asyncio
from asgiref.sync import sync_to_async
from typing import Dict, Optional, Union, List, Tuple
from decimal import Decimal

from django.core.management.base import BaseCommand
from django.db.models import Q
from stellar_sdk.exceptions import NotFoundError
from stellar_sdk.transaction import Transaction as HorizonTransaction
from stellar_sdk.transaction_envelope import TransactionEnvelope
from stellar_sdk.xdr import (
    PaymentResult,
    PathPaymentStrictSendResult,
    PathPaymentStrictReceiveResult,
    OperationResult,
    TransactionResult,
)
from stellar_sdk.xdr.utils import from_xdr_amount
from stellar_sdk.operation import (
    Operation,
    Payment,
    PathPaymentStrictReceive,
    PathPaymentStrictSend,
)
from stellar_sdk.server import Server
from stellar_sdk.client.aiohttp_client import AiohttpClient

from polaris import settings
from polaris.models import Asset, Transaction
from polaris.utils import getLogger, maybe_make_callback_async

logger = getLogger(__name__)
PaymentOpResult = Union[
    PaymentResult, PathPaymentStrictSendResult, PathPaymentStrictReceiveResult
]
PaymentOp = Union[Payment, PathPaymentStrictReceive, PathPaymentStrictSend]


class Command(BaseCommand):
    """
    Streams transactions for the distribution account of each Asset in the DB.

    For every response from the server, attempts to find a matching transaction in
    the database with `find_matching_payment_op` and updates the transaction's
    status to `pending_anchor` or `pending_receiver` depending on the protocol.

    Then, the ``execute_outgoing_transaction`` process will query for transactions
    in those statuses and provide the anchor an integration function for executing
    the payment or withdrawal.
    """

    def handle(self, *_args, **_options):  # pragma: no cover
        try:
            asyncio.run(self.watch_transactions())
        except Exception as e:
            # This is very likely a bug, so re-raise the error and crash.
            # Heroku will restart the process unless it is repeatedly crashing,
            # in which case restarting isn't of much use.
            logger.exception("watch_transactions() threw an unexpected exception")
            raise e

    async def watch_transactions(self):  # pragma: no cover
        assets = await sync_to_async(list)(
            Asset.objects.exclude(distribution_seed__isnull=True)
        )
        await asyncio.gather(
            *[self._for_account(asset.distribution_account) for asset in assets]
        )

    async def _for_account(self, account: str):
        """
        Stream transactions for the server Stellar address.
        """
        async with Server(settings.HORIZON_URI, client=AiohttpClient()) as server:
            try:
                # Ensure the distribution account actually exists
                await server.load_account(account)
            except NotFoundError:
                # This exception will crash the process, but the anchor needs
                # to provide valid accounts to watch.
                raise RuntimeError(
                    "Stellar distribution account does not exist in horizon"
                )
            last_completed_transaction = await sync_to_async(
                Transaction.objects.filter(
                    Q(kind=Transaction.KIND.withdrawal) | Q(kind=Transaction.KIND.send),
                    receiving_anchor_account=account,
                    status=Transaction.STATUS.completed,
                )
                .order_by("-completed_at")
                .first
            )()

            cursor = "0"
            if last_completed_transaction:
                cursor = last_completed_transaction.paging_token

            logger.info(
                f"starting transaction stream for {account} with cursor {cursor}"
            )
            endpoint = server.transactions().for_account(account).cursor(cursor)
            async for response in endpoint.stream():
                await self.process_response(response, account)

    @classmethod
    async def process_response(cls, response, account):
        # We should not match valid pending transactions with ones that were
        # unsuccessful on the stellar network. If they were unsuccessful, the
        # client is also aware of the failure and will likely attempt to
        # resubmit it, in which case we should match the resubmitted transaction
        if not response.get("successful"):
            return

        try:
            _ = response["id"]
            envelope_xdr = response["envelope_xdr"]
            memo = response["memo"]
            result_xdr = response["result_xdr"]
        except KeyError:
            return

        # Query filters for SEP6 and 24

        withdraw_filters = Q(
            status=Transaction.STATUS.pending_user_transfer_start,
            kind__in=[
                Transaction.KIND.withdrawal,
                getattr(Transaction.KIND, "withdrawal-exchange"),
            ],
        )
        # Query filters for SEP31
        send_filters = Q(
            status=Transaction.STATUS.pending_sender, kind=Transaction.KIND.send,
        )
        transactions = await sync_to_async(list)(
            Transaction.objects.filter(
                withdraw_filters | send_filters,
                memo=memo,
                receiving_anchor_account=account,
            )
            .select_related("asset")
            .all()
        )
        if not transactions:
            logger.info(f"No match found for stellar transaction {response['id']}")
            return
        elif len(transactions) == 1:
            transaction = transactions[0]
        else:
            # in the prior implementation of watch_transactions, the first transaction
            # to have the same memo is matched, so we'll do the same in the refactored
            # version.
            logger.error(f"multiple Transaction objects returned for memo: {memo}")
            transaction = transactions[0]

        logger.info(
            f"Matched transaction object {transaction.id} for stellar transaction {response['id']}"
        )

        op_results = TransactionResult.from_xdr(result_xdr).result.results
        horizon_tx = TransactionEnvelope.from_xdr(
            envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
        ).transaction

        payment_data = await cls._find_matching_payment_data(
            response, horizon_tx, op_results, transaction
        )
        if not payment_data:
            logger.warning(f"Transaction matching memo {memo} has no payment operation")
            return

        # Transaction.amount_in is overwritten with the actual amount sent in the stellar
        # transaction. This allows anchors to validate the actual amount sent in
        # execute_outgoing_transactions() and handle invalid amounts appropriately.
        transaction.amount_in = round(
            Decimal(payment_data["amount"]), transaction.asset.significant_decimals,
        )

        # The stellar transaction has been matched with an existing record in the DB.
        # Now the anchor needs to initiate the off-chain transfer of the asset.
        if transaction.protocol == Transaction.PROTOCOL.sep31:
            # SEP-31 uses 'pending_receiver' status
            transaction.status = Transaction.STATUS.pending_receiver
            await sync_to_async(transaction.save)()
        else:
            # SEP-6 and 24 uses 'pending_anchor' status
            transaction.status = Transaction.STATUS.pending_anchor
            await sync_to_async(transaction.save)()
        await maybe_make_callback_async(transaction)
        return

    @classmethod
    async def _find_matching_payment_data(
        cls,
        response: Dict,
        horizon_tx: HorizonTransaction,
        result_ops: List[OperationResult],
        transaction: Transaction,
    ) -> Optional[Dict]:
        matching_payment_data = None
        ops = horizon_tx.operations
        for idx, op_result in enumerate(result_ops):
            op, op_result = cls._cast_operation_and_result(ops[idx], op_result)
            if not op_result:  # not a payment op
                continue
            maybe_payment_data = cls._check_for_payment_match(
                op, op_result, transaction.asset
            )
            if maybe_payment_data:
                if ops[idx].source:
                    source = ops[idx].source.account_muxed or ops[idx].source.account_id
                else:
                    source = (
                        horizon_tx.source.account_muxed or horizon_tx.source.account_id
                    )
                await cls._update_transaction_info(
                    transaction, response["id"], response["paging_token"], source
                )
                matching_payment_data = maybe_payment_data
                break

        return matching_payment_data

    @classmethod
    async def _update_transaction_info(
        cls, transaction: Transaction, stellar_txid: str, paging_token: str, source: str
    ):
        transaction.stellar_transaction_id = stellar_txid
        transaction.from_address = source
        transaction.paging_token = paging_token
        await sync_to_async(transaction.save)()

    @classmethod
    def _check_for_payment_match(
        cls, operation: PaymentOp, op_result: PaymentOpResult, want_asset: Asset
    ) -> Optional[Dict]:
        payment_data = cls._get_payment_values(operation, op_result)
        if (
            payment_data["destination"] == want_asset.distribution_account
            and payment_data["code"] == want_asset.code
            and payment_data["issuer"] == want_asset.issuer
        ):
            return payment_data
        else:
            return None

    @classmethod
    def _cast_operation_and_result(
        cls, operation: Operation, op_result: OperationResult
    ) -> Tuple[Optional[PaymentOp], Optional[PaymentOpResult]]:
        op_xdr_obj = operation.to_xdr_object()
        if isinstance(operation, Payment):
            return (
                Payment.from_xdr_object(op_xdr_obj),
                op_result.tr.payment_result,
            )
        elif isinstance(operation, PathPaymentStrictSend):
            return (
                PathPaymentStrictSend.from_xdr_object(op_xdr_obj),
                op_result.tr.path_payment_strict_send_result,
            )
        elif isinstance(operation, PathPaymentStrictReceive):
            return (
                PathPaymentStrictReceive.from_xdr_object(op_xdr_obj),
                op_result.tr.path_payment_strict_receive_result,
            )
        else:
            return None, None

    @classmethod
    def _get_payment_values(
        cls, operation: PaymentOp, op_result: PaymentOpResult
    ) -> Dict:
        values = {
            "destination": operation.destination.account_id,
            "amount": None,
            "code": None,
            "issuer": None,
        }
        if isinstance(operation, Payment):
            values["amount"] = str(operation.amount)
            values["code"] = operation.asset.code
            values["issuer"] = operation.asset.issuer
        elif isinstance(operation, PathPaymentStrictSend):
            # since the dest amount is not specified in a strict-send op,
            # we need to get the dest amount from the operation's result
            #
            # this method of fetching amounts gives the "raw" amount, so
            # we need to divide by Operation._ONE: 10000000
            # (Stellar uses 7 decimals places of precision)
            values["amount"] = from_xdr_amount(op_result.success.last.amount.int64)
            values["code"] = operation.dest_asset.code
            values["issuer"] = operation.dest_asset.issuer
        elif isinstance(operation, PathPaymentStrictReceive):
            values["amount"] = str(operation.dest_amount)
            values["code"] = operation.dest_asset.code
            values["issuer"] = operation.dest_asset.issuer
        else:
            raise ValueError("Unexpected operation, expected payment or path payment")
        return values
